package main

import (
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"io"
	"io/ioutil"
	"net/http"
	"os"
	"strings"
	"sync"

	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe/mmap"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/probe/self"
	"gitlab.com/gitlab-org/gitlab-metrics-exporter/internal/render"

	goLog "log"

	"gitlab.com/gitlab-org/labkit/log"
	"gitlab.com/gitlab-org/labkit/metrics"

	"github.com/urfave/cli/v2"
)

const (
	metricsEndpoint  = "/metrics"
	metricsNamespace = "gitlab_metrics_exporter"

	probeSelf = "self"
	probeMmap = "mmap"

	defaultPort      = "8082"
	defaultLogFile   = "stderr"
	defaultLogFormat = "text"
	defaultLogLevel  = "info"
)

var (
	defaultProbes   = []string{probeSelf}
	availableProbes = []string{probeSelf, probeMmap}
)

type config struct {
	host      string
	port      string
	certFile  string
	certKey   string
	probes    []probe.Probe
	logFile   string
	logLevel  string
	logFormat string
}

func main() {
	err := buildCliApp().Run(os.Args)
	if err != nil {
		goLog.Fatalf("server terminated: %v", err)
	}
}

func buildCliApp() *cli.App {
	return &cli.App{
		Usage: "A Prometheus exporter for GitLab application metrics",
		Flags: buildCliFlags(),
		Action: func(c *cli.Context) error {
			cfg, err := loadConfig(c)
			if err != nil {
				return err
			}
			return startServer(&cfg)
		},
	}
}

func buildCliFlags() []cli.Flag {
	sharedFlags := []cli.Flag{
		&cli.StringFlag{
			Name:    "host",
			Usage:   "TCP server address host",
			EnvVars: []string{"GME_SERVER_HOST"},
		},
		&cli.StringFlag{
			Name:    "port",
			Value:   defaultPort,
			Usage:   "TCP server address port",
			EnvVars: []string{"GME_SERVER_PORT"},
		},
		&cli.StringFlag{
			Name:    "cert-file",
			Usage:   "Path to TLS certificate PEM file (combined format); requires --cert-key to be set",
			EnvVars: []string{"GME_CERT_FILE"},
		},
		&cli.StringFlag{
			Name:    "cert-key",
			Usage:   "Path to TLS certificate key; requires --cert-file to be set",
			EnvVars: []string{"GME_CERT_KEY"},
		},
		&cli.StringSliceFlag{
			Name:    "probe",
			Value:   cli.NewStringSlice(defaultProbes...),
			Usage:   "probe to run when metrics are requested (available: " + strings.Join(availableProbes, ", ") + ")",
			EnvVars: []string{"GME_PROBES"},
			Aliases: []string{"p"},
		},
		&cli.StringFlag{
			Name:    "log-file",
			Value:   defaultLogFile,
			Usage:   "Path to a file receiving log output (including stdout, stderr)",
			EnvVars: []string{"GME_LOG_FILE"},
		},
		&cli.StringFlag{
			Name:    "log-format",
			Value:   defaultLogFormat,
			Usage:   "One of text, color, json",
			EnvVars: []string{"GME_LOG_FORMAT"},
		},
		&cli.StringFlag{
			Name:    "log-level",
			Value:   defaultLogLevel,
			Usage:   "One of quiet, panic, fatal, error, warn, info, debug, trace",
			EnvVars: []string{"GME_LOG_LEVEL"},
		},
	}

	return append(sharedFlags, mmapProbeFlags()...)
}

func mmapProbeFlags() []cli.Flag {
	return []cli.Flag{
		&cli.StringFlag{
			Name:    "mmapdir",
			Usage:   "mmap probe: absolute path to prometheus-client-mmap metrics files",
			EnvVars: []string{"GME_MMAP_METRICS_DIR"},
		},
	}
}

func loadConfig(c *cli.Context) (config, error) {
	cfg := config{
		host:      c.String("host"),
		port:      c.String("port"),
		certFile:  c.String("cert-file"),
		certKey:   c.String("cert-key"),
		logFile:   c.String("log-file"),
		logFormat: c.String("log-format"),
		logLevel:  c.String("log-level"),
	}

	probes, err := loadProbes(c)
	if err != nil {
		return cfg, err
	}
	cfg.probes = probes

	return cfg, nil
}

func loadProbes(c *cli.Context) ([]probe.Probe, error) {
	probeNames := c.StringSlice("probe")
	var probes []probe.Probe
	for _, probeName := range probeNames {
		if !isSupportedProbe(probeName) {
			return nil, fmt.Errorf("unsupported probe: %s", probeName)
		}

		switch probeName {
		case "self":
			probes = append(probes, self.NewProbe())
		case "mmap":
			metricsDir := c.String("mmapdir")
			if metricsDir == "" {
				return nil, fmt.Errorf("cannot be blank: --mmapdir")
			}
			probes = append(probes, mmap.NewDefaultProbe(metricsDir))
		}
	}

	return probes, nil
}

func isSupportedProbe(probe string) bool {
	for _, p := range availableProbes {
		if p == probe {
			return true
		}
	}
	return false
}

func startServer(cfg *config) error {
	logCloser, err := internal.SetupLog(cfg.logFile, cfg.logFormat, cfg.logLevel)
	if err != nil {
		return err
	}
	defer logCloser.Close()

	http.Handle(metricsEndpoint,
		log.AccessLogger(
			metrics.NewHandlerFactory(metrics.WithLabels("path"), metrics.WithNamespace(metricsNamespace))(
				http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {
					if err := writeProbeResults(cfg.probes, rw); err != nil {
						rw.WriteHeader(500)
					}
				}),
				metrics.WithLabelValues(map[string]string{"path": metricsEndpoint}),
			)))

	srvAddr := cfg.host + ":" + cfg.port
	log.WithField("address", srvAddr).WithField("pid", os.Getpid()).Info("server start")

	if cfg.certFile != "" && cfg.certKey != "" {
		caCert, err := ioutil.ReadFile(cfg.certFile)
		if err != nil {
			return fmt.Errorf("certificate file: %v", err)
		}
		caCertPool := x509.NewCertPool()
		caCertPool.AppendCertsFromPEM(caCert)

		tlsConfig := &tls.Config{
			ClientCAs:  caCertPool,
			ClientAuth: tls.RequireAndVerifyClientCert,
			MinVersion: tls.VersionTLS12,
		}

		server := http.Server{
			Addr:      srvAddr,
			TLSConfig: tlsConfig,
		}

		return server.ListenAndServeTLS(cfg.certFile, cfg.certKey)
	}

	return http.ListenAndServe(srvAddr, nil)
}

func writeProbeResults(probes []probe.Probe, w io.Writer) error {
	ch := make(chan probe.ProbeResult)

	wg := sync.WaitGroup{}

	for _, p := range probes {
		wg.Add(1)

		p_ := p
		go func() {
			defer wg.Done()
			if err := p_.Probe(ch); err != nil {
				log.WithField("probe", p_.String()).Errorf("Probe: %v", err)
			}
		}()
	}

	go func() {
		defer close(ch)
		wg.Wait()
	}()

	return render.NewRenderer().TextAsync(ch, w)
}
